@@ -1,4 +1,4 @@
-# pylint: disable-msg=R0903
+# pylint: disable=R0903
 """test class members"""
 
 __revision__ = ''
@@ -14,6 +14,7 @@ class MyClass:
         """test"""
         self.correct += 2
         self.incorrect += 2
+        del self.havenot
         self.nonexistent1.truc()
         self.nonexistent2[1] = 'hehe'
 
